#include "mpi.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

/*
extern double drand48();
*/

/* Pipeline version of the algorithm... */
/* we really need the velocities as well... */
typedef struct {
    double x, y, z;
    double mass;
    } Particle;
/* We use leapfrog for the time integration ... */
typedef struct {
    double xold, yold, zold;
    double fx, fy, fz;
    } ParticleV;


void InitParticles( Particle[], ParticleV [], unsigned int, unsigned int );
double ComputeForces( Particle [], ParticleV [], unsigned int, unsigned int, unsigned int );
double ComputeNewPos( Particle [], ParticleV [], unsigned int, unsigned int, double, MPI_Comm );

#define MAX_PARTICLES 4000
#define DEFAULT_TIME_STEPS 10

main( int argc, char *argv[] )
{
    Particle  particles[MAX_PARTICLES];   /* Particles on ALL nodes */
    ParticleV pv[MAX_PARTICLES];          /* Particle velocity */
    Particle  sendbuf[MAX_PARTICLES];     /* buffers */
	//recvbuf[MAX_PARTICLES];
    //MPI_Request request[2];
    int rank, size, npart;
    unsigned int i,
	             offset,               /* location of local particles */
                 totpart,              /* total number of particles */
                 cnt;                  /* number of times in loop */
    MPI_Datatype particletype;
    double      sim_t;                      /* Simulation time */
    double      time;                       /* Computation time */
    //MPI_Comm    group_comm;
    //MPI_Status  statuses[2];

    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    MPI_Type_contiguous( 4, MPI_DOUBLE, &particletype );
    MPI_Type_commit( &particletype );

    if (argc < 2) {
    	cnt = DEFAULT_TIME_STEPS;
    }
    else {
    	cnt = atoi(argv[1]);
    }
    if (argc < 3) {
    	totpart = MAX_PARTICLES;
    }
    else {
    	totpart = atoi(argv[2]);
    }

    npart = totpart/size;
    if (totpart % size != 0) {
    	npart++;
    	totpart = npart * size;
    }

    offset = rank * npart;

    if (totpart > MAX_PARTICLES) {
    	fprintf(stderr,"Error: Too much particles.\n");
    	MPI_Finalize();
    	return 1;
    }

/* Generate the initial values */
    memset (particles, 0 , totpart*sizeof(Particle));
    InitParticles( particles, pv, offset, npart);

    time = MPI_Wtime();
    sim_t = 0.0;
    while (cnt--) {
		double max_f, max_f_seg;
		MPI_Allgather( &particles[offset], npart, particletype, particles, npart, particletype, MPI_COMM_WORLD);

		max_f = 0.0;

    	/* Compute forces (2D only) */
    	max_f_seg = ComputeForces( particles, pv, totpart, offset, npart );
		if (max_f_seg > max_f) max_f = max_f_seg;

		/* Once we have the forces, we compute the changes in position */
		sim_t += ComputeNewPos( particles, pv, offset, npart, max_f, MPI_COMM_WORLD );

		/* We could do graphics here (move particles on the display) */
    } //while
    time = MPI_Wtime() - time;
    if (rank == 0) {
		printf( "Computed %d particles in %f seconds\n", totpart, time );
    } //if
    MPI_Finalize();
    return 0;
}

void InitParticles( Particle particles[], ParticleV pv[], unsigned int offset, unsigned int npart )
{
    unsigned int i;
    for (i=offset; i<offset+npart; i++) {
		particles[i].x	  = drand48();
		particles[i].y	  = drand48();
		particles[i].z	  = drand48();
		particles[i].mass = 1.0;
		pv[i].xold	  = particles[i].x;
		pv[i].yold	  = particles[i].y;
		pv[i].zold	  = particles[i].z;
		pv[i].fx	  = 0;
		pv[i].fy	  = 0;
		pv[i].fz	  = 0;
    }
}

double ComputeForces( Particle particles[], ParticleV pv[], unsigned int totpart,
					unsigned int offset, unsigned int npart )
{
  double max_f, rmin;
  int i, j;

  max_f = 0.0;
  for (i=offset; i<offset+npart; i++) {
    double xi, yi, rx, ry, mj, r, fx, fy;
    rmin = 100.0;
    xi   = particles[i].x;
    yi   = particles[i].y;
    fx   = 0.0;
    fy   = 0.0;
    for (j=0; j<totpart; j++) {
      rx = xi - particles[j].x;
      ry = yi - particles[j].y;
      mj = particles[j].mass;
      r  = rx * rx + ry * ry;
      /* ignore overlap and same particle */
      if (r == 0.0) continue;
      if (r < rmin) rmin = r;
      /* compute forces */
      r  = r * sqrt(r);
      fx -= mj * rx / r;
      fy -= mj * ry / r;
    }
    pv[i].fx += fx;
    pv[i].fy += fy;
    /* Compute a rough estimate of (1/m)|df / dx| */
    fx = sqrt(fx*fx + fy*fy)/rmin;
    if (fx > max_f) max_f = fx;
  }
  return max_f;
}

double ComputeNewPos( Particle particles[], ParticleV pv[], unsigned int offset, unsigned int npart,
				double max_f, MPI_Comm comm )
{
  int i;
  double      a0, a1, a2;
  static      double dt_old = 0.001, dt = 0.001;
  double      dt_est, dt_new;

  /* integration is a0 * x^+ + a1 * x + a2 * x^- = f / m */
  a0	 = 2.0 / (dt * (dt + dt_old));
  a2	 = 2.0 / (dt_old * (dt + dt_old));
  a1	 = -(a0 + a2);      /* also -2/(dt*dt_old) */

  for (i=offset; i<offset+npart; i++) {
    double xi, yi;
    /* Very, very simple leapfrog time integration.  We use a variable
       step version to simplify time-step control. */
    xi	           = particles[i].x;
    yi	           = particles[i].y;
    particles[i].x = (pv[i].fx - a1 * xi - a2 * pv[i].xold) / a0;
    particles[i].y = (pv[i].fy - a1 * yi - a2 * pv[i].yold) / a0;
    pv[i].xold     = xi;
    pv[i].yold     = yi;
    pv[i].fx       = 0;
    pv[i].fy       = 0;
  }

  /* Recompute a time step. Stability criteria is roughly
     2/sqrt(1/m |df/dx|) >= dt.  We leave a little room */
  dt_est = 1.0/sqrt(max_f);
  /* Set a minimum: */
  if (dt_est < 1.0e-6) dt_est = 1.0e-6;
  MPI_Allreduce( &dt_est, &dt_new, 1, MPI_DOUBLE, MPI_MIN, comm );
  /* Modify time step */
  if (dt_new < dt) {
    dt_old = dt;
    dt     = dt_new;
  }
  else if (dt_new > 4.0 * dt) {
    dt_old = dt;
    dt    *= 2.0;
  }

  return dt_old;
}
